@@ -12,9 +12,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+    You should have received a copy of the GNU General Public License along
+    with this program; if not, write to the Free Software Foundation, Inc.,
+    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
 #include "config.h"
@@ -198,7 +198,7 @@ static gint About_clicked(
     gtk_show_about_dialog(GTK_WINDOW(main_window)
                           , "version", PACKAGE_VERSION, "copyright",
                           "Copyright \xc2\xa9 1997,1998  Matt Kimball",
-                          "website", "http://www.bitwizard.nl/mtr/",
+                          "website", "https://www.bitwizard.nl/mtr/",
                           "authors", authors, "comments",
                           "The 'traceroute' and 'ping' programs in a single network diagnostic tool.",
                           "license",
@@ -294,8 +294,7 @@ static void Toolbar_fill(
 
     /* allow root only to set zero delay */
     Adjustment = (GtkAdjustment *) gtk_adjustment_new(ctl->WaitTime,
-                                                      getuid() ==
-                                                      0 ? 0.01 : 1.00,
+                                                      running_as_root() ? 0.01 : 1.00,
                                                       999.99, 1.0, 10.0,
                                                       0.0);
     Button = gtk_spin_button_new(Adjustment, 0.5, 2);
